--  dh_ada_library, helper for Ada libraries Debian maintainers
--
--  Copyright (C) 2012 Nicolas Boulenguez <nicolas.boulenguez@free.fr>
--
--  This program is free software: you can redistribute it and/or
--  modify it under the terms of the GNU General Public License as
--  published by the Free Software Foundation, either version 3 of the
--  License, or (at your option) any later version.
--  This program is distributed in the hope that it will be useful, but
--  WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
--  General Public License for more details.
--  You should have received a copy of the GNU General Public License
--  along with this program. If not, see <http://www.gnu.org/licenses/>.
--
with Ada.Characters.Handling;
with Ada.Command_Line;
with Ada.Containers.Indefinite_Vectors;
with Ada.Directories;
with Ada.Strings.Fixed;
with Ada.Strings.Maps;
with Ada.Text_IO;

with GNAT.OS_Lib;
with GNAT.Regpat;
with Csets;                             --  gnatvsn
with Namet;                             --  gnatvsn
with Snames;                            --  gnatvsn
with Prj.Pars;
with Prj.Conf;
with Prj.Tree;
with Prj.Util;

with Dh_Lib;
with Environment_Variable_Backups;
with Misc;
with String_Sets;

procedure Dh_Ada_Library is

   --  It would be natural to process each package in parallel, but
   --  Prj relies on the filesystem state (Prj configuration is
   --  created if absent for example), and some dh_tools have side
   --  effects on other packages (dh_strip for example).

   ----------------------------------------------------------------------
   use type GNAT.Regpat.Match_Location;
   use type Namet.Path_Name_Type;
   use type Prj.Lib_Kind;
   use type Prj.Project_Id;
   use type Prj.Project_List;
   use type Prj.Source_Id;
   use type Prj.String_List_Id;
   use type Prj.Tree.Project_Node_Tree_Ref;
   use type Prj.Variable_Kind;
   use type String_Sets.Cursor;

   Consistency_Error   : exception;
   Project_Parse_Error : exception;
   Usage_Error         : exception;
   ----------------------------------------------------------------------
   --  String constants

   Minus_Dumpversion : aliased String := "-dumpversion";
   Gnatgcc_Version : constant String := Misc.Output_Of
     ("/usr/bin/gnatgcc",
      (1 => Minus_Dumpversion'Unchecked_Access));

   Minus_Q_Deb_Host_Multiarch : aliased String := "-qDEB_HOST_MULTIARCH";
   Deb_Host_Multiarch  : constant String := Misc.Output_Of
     ("/usr/bin/dpkg-architecture",
      (1 =>  Minus_Q_Deb_Host_Multiarch'Unchecked_Access));

   Generated_By_Header : constant String
     := "(generated by " & Ada.Command_Line.Command_Name
     & " during package build)";

   Space_Character_Set : constant Ada.Strings.Maps.Character_Set
     := Ada.Strings.Maps.To_Set (' ');
   Configuration_File  : constant String := "debian/ada_libraries";

   Src_Install_Dir     : constant String := "usr/share/ada/adainclude/";
   Ali_Install_Dir     : constant String := "usr/lib/" & Deb_Host_Multiarch
     & "/ada/adalib/";
   Lib_Install_Dir     : constant String := "usr/lib/" & Deb_Host_Multiarch
     & '/';

   ----------------------------------------------------------------------
   --  Types and procedures

   type Affectation_Argument
     (Name_Length  : Natural;
      Value_Length : Natural) is record
      Name         : String (1 .. Name_Length);
      Value        : String (1 .. Value_Length);
   end record;

   pragma Warnings (Off,
                    "pragma Warnings Off for ""*"" unused, could be omitted");

   package Affectation_Argument_Vectors is
      new Ada.Containers.Indefinite_Vectors (Positive, Affectation_Argument);

   pragma Warnings (On,
                    "pragma Warnings Off for ""*"" unused, could be omitted");

   type Project_Argument_Group
     (Path_Length  : Natural) is record
      Path         : String (1 .. Path_Length); --  including ".gpr" suffix
      Affectations : Affectation_Argument_Vectors.Vector;
   end record;

   pragma Warnings (Off,
                    "pragma Warnings Off for ""*"" unused, could be omitted");

   package Project_Argument_Group_Vectors is
      new Ada.Containers.Indefinite_Vectors
     (Positive, Project_Argument_Group);

   pragma Warnings (On,
                    "pragma Warnings Off for ""*"" unused, could be omitted");


   function Package_Containing
     (Pattern : not null GNAT.OS_Lib.String_Access)
     return String;
   --  Wrapper to a dpkg-query call. "" is returned if there is no
   --  match or many matches.


   procedure Parse_Arguments
     (Projects  : in out Project_Argument_Group_Vectors.Vector;
      Processed :    out Dh_Lib.Command_Line_Argument_Set);

   procedure List_Packages_In_Debian_Control
     (Packages_In_Control : out String_Sets.Set);

   procedure Parse_Project
     (Project          : in       Project_Argument_Group;
      Project_Tree_Ref : not null Prj.Project_Tree_Ref;
      Project_Id       :      out Prj.Project_Id);
   procedure Process_Library
     (Project_Tree_Ref    : not null Prj.Project_Tree_Ref;
      Project_Id          : not null Prj.Project_Id;
      Packages_In_Control : in       String_Sets.Set);

   --  Helpers for Process_Project

   procedure Install_Project
     (Project_Tree_Ref : not null Prj.Project_Tree_Ref;
      Project_Id       : not null Prj.Project_Id;
      Dev_Pkg_Dir      : in       String);
   procedure Install_Ali_Files
     (Project_Id  : not null Prj.Project_Id;
      Dev_Pkg_Dir : in       String);
   procedure Add_Substitution_Variables_For_Withed_Projects
     (Project_Id          : not null Prj.Project_Id;
      Dev_Pkg             : in       String;
      Packages_In_Control : in       String_Sets.Set);
   --  For each project imported by the build project, try to figure
   --  where it comes from. If the name matches a -dev package built
   --  from the same source, add a Depends. Else, if it lies in the
   --  standard directory for system-wide project, use dpkg to find
   --  the Debian -dev package providing it. Else, output a warning.
   procedure Append_Trampoline_Lintian_Override
     (Project_Id  : not null Prj.Project_Id;
      Lib_Pkg     : in       String;
      Lib_Pkg_Dir : in       String;
      Soname      : in       String);
   --  If the shared library declares its stack executable, which is
   --  not recommeded by the policy, add a lintian override explaining
   --  that GNAT handles exceptions with trampolines. We check the
   --  freshly built lib, since if Noact is set it may not be
   --  installed yet.

   --  We must call dh_lintian before writing to the file. If run
   --  later, it may overwrite overwrite our changes.
   procedure Install_Shared_Library
     (Project_Id  : not null Prj.Project_Id;
      Lib_Pkg_Dir : in       String;
      Soname      : in       String);
   procedure Install_Shared_Object_Symbolic_Link
     (Project_Id  : not null Prj.Project_Id;
      Dev_Pkg_Dir : in       String;
      Soname      : not null GNAT.OS_Lib.String_Access);
   procedure Install_Sources
     (Project_Tree_Ref : not null Prj.Project_Tree_Ref;
      Project_Id       : not null Prj.Project_Id;
      Dev_Pkg_Dir      : in       String);
   procedure Install_Static_Library
     (Project_Id  : not null Prj.Project_Id;
      Dev_Pkg_Dir : in       String);
   --  Guess where the freshly built static archive lies, then install
   --  it. A warning is displayed if there is not exactly one find
   --  named "libNAME.a".
   procedure Run_Dh_Installdocs
     (Pkg      : in String;
      Link_Doc : in String);
   procedure Run_Dh_Strip_Excluding_Static_Archive
     (Project_Id : not null Prj.Project_Id;
      Dev_Pkg    : in       String);
   procedure Run_Dh_Strip_For_Shared_Library
     (Lib_Pkg : in String;
      Dbg_Pkg : in String);

   --  Small inlined routines.

   function Lib_Name (Project_Id : not null Prj.Project_Id) return String;
   pragma Inline (Lib_Name);

   function With_Minus_If_Needed (Name : in String) return String;
   pragma Inline (With_Minus_If_Needed);

   function Find_Package
     (Packages_In_Control : in String_Sets.Set;
      Pattern             : in String)
     return String;
   --  If no element of Packages_In_Control matches Pattern, "" is
   --  returned. Else, the substring matching the first subexpression
   --  of the first matching element is returned.

   --  Constraint_Error is raisde if Pattern does not include at least
   --  a subgroup.

   ----------------------------------------------------------------------
   --  Implementation

   procedure Add_Substitution_Variables_For_Withed_Projects
     (Project_Id          : not null Prj.Project_Id;
      Dev_Pkg             : in       String;
      Packages_In_Control : in       String_Sets.Set)
   is
      Project_List : Prj.Project_List := Project_Id.all.Imported_Projects;
   begin
      while Project_List /= null loop
         declare
            use String_Sets;
            Withed_Prj : constant Prj.Project_Id := Project_List.all.Project;
            Withed_Name : constant String := Lib_Name (Withed_Prj);
            Dev_Pattern : constant String := "^(lib"
              & With_Minus_If_Needed (Withed_Name) & "[[:digit:].]+-dev)$";
            Dev_Control : constant String
              := Find_Package (Packages_In_Control, Dev_Pattern);
         begin
            if Dev_Control /= "" then
               --  Withed library is built from same source package.
               Dh_Lib.Addsubstvar
                 (Binary_Package                   => Dev_Pkg,
                  Substvar                         => "ada:Depends",
                  Package_That_Will_Be_Depended_On => Dev_Control,
                  Version_Info                     => "= ${binary:Version}");
            else
               declare
                  Withed_Path : aliased String
                    := Namet.Get_Name_String (Withed_Prj.all.Path.Name);
                  Dev_Dpkg_Query : constant String
                    := Package_Containing (Withed_Path'Unchecked_Access);
               begin
                  if Dev_Dpkg_Query /= "" then
                     Dh_Lib.Addsubstvar
                       (Binary_Package                   => Dev_Pkg,
                        Substvar                         => "ada:Depends",
                        Package_That_Will_Be_Depended_On => Dev_Dpkg_Query);
                  else
                     Dh_Lib.Warning
                       (Namet.Get_Name_String (Project_Id.all.Path.Name)
                          & " imports " & Withed_Path
                          & " not provided by same source or any installed"
                          & " package. No ada:Depends added.");
                  end if;
               end;
            end if;
         end;
         Project_List := Project_List.all.Next;
      end loop;
   end Add_Substitution_Variables_For_Withed_Projects;

   procedure Append_Trampoline_Lintian_Override
     (Project_Id  : not null Prj.Project_Id;
      Lib_Pkg     : in       String;
      Lib_Pkg_Dir : in       String;
      Soname      : in       String)
   is
      use Ada.Text_IO;
      use GNAT.OS_Lib;
      use GNAT.Regpat;
      Headers     : aliased String := "--program-headers";
      Wide        : aliased String := "--wide";
      So_Path     : aliased String
        := Namet.Get_Name_String (Project_Id.all.Library_Dir.Name) & Soname;
      Temp_File   : String_Access;
      F           : File_Type;
      Pkg_Equals  : aliased String := "--package=" & Lib_Pkg;
      Target_Dir  : constant String
        := Lib_Pkg_Dir & "usr/share/lintian/overrides/";
      Overrides   : constant String := Target_Dir & Lib_Pkg;
      Header      : constant String := "# " & Generated_By_Header;
      Regex       : constant Pattern_Matcher := Compile
        ("^[[:space:]]*GNU_STACK[[:space:]]+(0x[[:xdigit:]]+[[:space:]]+){5}"
           & "RW(E)?[[:space:]]+0x[[:xdigit:]]+$");
      MA            : Match_Array (0 .. Paren_Count (Regex));
      Stack_Declared_Executable : Boolean;
   begin
      Misc.Capture_Output ("/usr/bin/readelf",
                           (Headers'Unchecked_Access,
                            Wide'Unchecked_Access,
                            So_Path'Unchecked_Access), Temp_File);
      Open (F, In_File, Temp_File.all);
      loop
         if End_Of_File (F) then
            Dh_Lib.Warning ("failed to parse readelf output, "
                              & "no lintian override generated. "
                              & "Please report to"
                              & Ada.Command_Line.Command_Name & " author.");
            Stack_Declared_Executable := False;
            exit;
         end if;
         Match (Regex, Get_Line (F), MA);
         if MA (0) /= No_Match then
            Stack_Declared_Executable := MA (MA'Last) /= No_Match;
            exit;
         end if;
      end loop;
      Close (F);
      Ada.Directories.Delete_File (Temp_File.all);
      Free (Temp_File);
      if not Stack_Declared_Executable then
         return;
      end if;
      Dh_Lib.Do_It ("/usr/bin/dh_lintian",
                    (1 => Pkg_Equals'Unchecked_Access));
      Dh_Lib.Create_Path (Target_Dir);
      Dh_Lib.Verbose_Print ("Adding executable stack lintian override to "
                              & Overrides & ".");
      if Dh_Lib.Noact then
         return;
      end if;
      if Ada.Directories.Exists (Overrides) then
         if Misc.A_Line_Matches (Overrides, '^' & Quote (Header) & '$') then
            return;
         end if;
         Open (F, Append_File, Overrides);
         New_Line (F);
      else
         Create (F, Out_File, Overrides);
      end if;
      Put_Line (F, Header);
      Put_Line (F, "# GNAT uses trampolines to implement Ada exceptions.");
      Put_Line (F, "# The wildcard replaces an architecture-dependent path,");
      Put_Line (F, "# allowing this file to be installed in /usr/share.");
      Put_Line (F, Lib_Pkg & " binary: shlib-with-executable-stack usr/lib/*/"
                  & Soname);
      Close (F);
   exception
      when others =>
         if Is_Open (F) then
            Close (F);
         end if;
         if Temp_File /= null then
            Ada.Directories.Delete_File (Temp_File.all);
            Free (Temp_File);
         end if;
         raise;
   end Append_Trampoline_Lintian_Override;

   function Find_Package
     (Packages_In_Control : in String_Sets.Set;
      Pattern             : in String)
     return String
   is
      use GNAT.Regpat;
      use String_Sets;
      PM  : constant Pattern_Matcher := Compile (Pattern);
      MA  : Match_Array (0 .. Paren_Count (PM));
      Pkg : Cursor := First (Packages_In_Control);
   begin
      while Pkg /= No_Element loop
         Match (PM, Element (Pkg), MA);
         if MA (0) /= No_Match then
            return Element (Pkg) (MA (1).First .. MA (1).Last);
         end if;
         Next (Pkg);
      end loop;
      return "";
   end Find_Package;

   procedure Install_Ali_Files
     (Project_Id  : not null Prj.Project_Id;
      Dev_Pkg_Dir : in       String)
   is
      use Ada.Directories;
      Target_Dir : constant String := Dev_Pkg_Dir &  Ali_Install_Dir
        & Lib_Name (Project_Id) & '/';
      procedure Process (ALI_File : in Directory_Entry_Type);
      procedure Process (ALI_File : in Directory_Entry_Type)
      is
      begin
         Dh_Lib.Copy_File (Full_Name (ALI_File),
                           Target_Dir & Simple_Name (ALI_File));
      end Process;
      Attr : Namet.Path_Name_Type := Project_Id.all.Library_ALI_Dir.Name;
   begin
      --  Optional Library_ALI_Dir attribute defaults to Library_Dir.
      if Attr = Namet.No_Path then
         Attr := Project_Id.all.Library_Dir.Name;
      end if;
      Dh_Lib.Create_Path (Target_Dir);
      Search (Namet.Get_Name_String (Attr), "*.ali",
              (Ordinary_File => True, others => False), Process'Access);
   end Install_Ali_Files;

   procedure Install_Project
     (Project_Tree_Ref : not null Prj.Project_Tree_Ref;
      Project_Id       : not null Prj.Project_Id;
      Dev_Pkg_Dir      : in       String)
   is
      use Ada.Text_IO, Ada;
      Target_Dir     : constant String := Dev_Pkg_Dir & Src_Install_Dir;
      File_Name      : constant String := Target_Dir & Lib_Name (Project_Id)
        & ".gpr";
      Lib_Opts_Attrs : constant array (Positive range <>) of Namet.Name_Id
        := (Snames.Name_Leading_Library_Options,
            Snames.Name_Library_Options);
      Project_List   : Prj.Project_List;
      Variable_Value : Prj.Variable_Value;
      String_List_Id : Prj.String_List_Id;
      F              : File_Type;
   begin
      Dh_Lib.Create_Path (Target_Dir);
      Dh_Lib.Verbose_Print ("Creating developer project " & File_Name);
      if Dh_Lib.Noact then return; end if;
      Create (F, Out_File, File_Name);
      Put_Line (F, "--  " & Generated_By_Header);
      Put_Line (F, "--  This project file is designed to help build "
                  & "applications that use " & Lib_Name (Project_Id) & '.');
      Put_Line (F, "--  Here is an example of how to use this project file:");
      Put_Line (F, "--");
      Put_Line (F, "--  with """ & Lib_Name (Project_Id) & """;");
      Put_Line (F, "--  project Example is");
      Put_Line (F, "--     for Object_Dir use ""obj"";");
      Put_Line (F, "--     for Exec_Dir use ""."";");
      Put_Line (F, "--     for Main use (""example"");");
      Put_Line (F, "--  end Example;");
      New_Line (F);

      --  Add same "with" import commands that were present in the
      --  original project.
      Project_List := Project_Id.all.Imported_Projects;
      while Project_List /= null loop
         if Project_List.all.Project.all.Library then
            Put_Line (F, "with """ & Lib_Name (Project_List.all.Project)
                        & """;");
         else
            Dh_Lib.Warning
              ("Imported project "
                 & Namet.Get_Name_String (Project_List.all.Project.all.Name)
                 & " is not a library, ignoring it.");
         end if;
         Project_List := Project_List.all.Next;
      end loop;

      Put_Line (F, "library project " & Lib_Name (Project_Id) & " is");
      Put_Line (F, "   for Library_Name use """ & Lib_Name (Project_Id)
                  & """;");
      Put_Line (F, "   for Library_Kind use ""dynamic"";");
      Put_Line (F, "   for Library_Dir use ""/" & Lib_Install_Dir & """;");
      Put_Line (F, "   for Source_Dirs use (""/" & Src_Install_Dir
                  & Lib_Name (Project_Id) & """);");
      Put_Line (F, "   for Library_ALI_Dir use ""/" & Ali_Install_Dir
                  & Lib_Name (Project_Id) & """;");
      Put_Line (F, "   for Externally_Built use ""True"";");

      --  Add a Linker package if -l or -L options were needed to
      --  build the package.

      --  Iterate on (Leading_)Library_Options attributes
      for Attribute_Name in Lib_Opts_Attrs'Range loop
         Variable_Value := Prj.Util.Value_Of
           (Variable_Name => Lib_Opts_Attrs (Attribute_Name),
            In_Variables  => Project_Id.all.Decl.Attributes,
            In_Tree       => Project_Tree_Ref);
         if not Variable_Value.Default then
            pragma Assert (Variable_Value.Project = Project_Id);
            pragma Assert (Variable_Value.Kind = Prj.List);

            --  Iterate on -l or -L options for this attribute.
            String_List_Id := Variable_Value.Values;
            while String_List_Id /= Prj.Nil_String loop
               declare
                  Option : constant String := Namet.Get_Name_String
                    (Project_Tree_Ref.all.String_Elements.Table.all
                       (String_List_Id).Value);
               begin
                  if Ada.Strings.Fixed.Head (Option, 2) = "-l"
                    or Ada.Strings.Fixed.Head (Option, 2) = "-L" then
                     if Col (F) = 1 then   --  first option
                        Put_Line (F, "   package Linker is");
                        Put (F, "      for Linker_Options use (");
                     elsif Col (F) + Option'Length + 3 > 80 then
                        --  line too long
                        Put_Line (F, ",");
                        Put (F, Ada.Strings.Fixed."*" (30, ' '));
                     else
                        Put (F, ", ");
                     end if;
                     Put (F, '"' & Option & '"');
                  end if;
               end;
               String_List_Id := Project_Tree_Ref.all.String_Elements.Table.all
                 (String_List_Id).Next;
            end loop;
            --  Finished with options in this attributes.

         end if;
      end loop;
      --  Finished with (Leading_)Library_Options attributes
      if Col (F) /= 1 then     --  at least a library option
         Put_Line (F, ");");
         Put_Line (F, "   end Linker;");
      end if;

      Put_Line (F, "end " & Lib_Name (Project_Id) & ";");
      Close (F);
   exception
      when others =>
         if Is_Open (F) then Close (F); end if;
         raise;
   end Install_Project;

   procedure Install_Shared_Library
     (Project_Id  : not null Prj.Project_Id;
      Lib_Pkg_Dir : in       String;
      Soname      : in       String)
   is
      Target_Dir : constant String := Lib_Pkg_Dir & Lib_Install_Dir;
   begin
      Dh_Lib.Create_Path (Target_Dir);
      Dh_Lib.Copy_File
        (Namet.Get_Name_String (Project_Id.all.Library_Dir.Name) & Soname,
         Target_Dir & Soname);
   end Install_Shared_Library;

   procedure Install_Shared_Object_Symbolic_Link
     (Project_Id  : not null Prj.Project_Id;
      Dev_Pkg_Dir : in       String;
      Soname      : not null GNAT.OS_Lib.String_Access)
   is
      Minus_Sf   : aliased String := "-sf";
      Target_Dir : constant String := Dev_Pkg_Dir & Lib_Install_Dir;
      Link_Name  : aliased String
        := Target_Dir & "lib" & Lib_Name (Project_Id) & ".so";
   begin
      Dh_Lib.Create_Path (Target_Dir);
      Dh_Lib.Do_It ("/bin/ln", (Minus_Sf'Unchecked_Access,
                                Soname,
                                Link_Name'Unchecked_Access));
   end Install_Shared_Object_Symbolic_Link;

   procedure Install_Sources
     (Project_Tree_Ref : not null Prj.Project_Tree_Ref;
      Project_Id       : not null Prj.Project_Id;
      Dev_Pkg_Dir      : in       String)
   is
      Target_Dir : constant String := Dev_Pkg_Dir & Src_Install_Dir
        & Lib_Name (Project_Id);
      Source_Iterator : Prj.Source_Iterator
        := Prj.For_Each_Source (Project_Tree_Ref, Project_Id);
      Source_Id : Prj.Source_Id;
   begin
      Dh_Lib.Create_Path (Target_Dir);
      loop
         Source_Id := Prj.Element (Source_Iterator);
         exit when Source_Id = Prj.No_Source;
         declare
            Full_Name : constant String
              := Namet.Get_Name_String (Source_Id.all.Path.Name);
         begin
            Dh_Lib.Copy_File
              (Full_Name,
               Target_Dir & '/' & Ada.Directories.Simple_Name (Full_Name));
         end;
         Prj.Next (Source_Iterator);
      end loop;
   end Install_Sources;

   procedure Install_Static_Library
     (Project_Id  : not null Prj.Project_Id;
      Dev_Pkg_Dir : in       String)
   is
      use Ada.Text_IO;
      use GNAT.OS_Lib;
      Target_Dir    : constant String := "./" & Dev_Pkg_Dir & Lib_Install_Dir;
      Simple_Name   : aliased String := "lib" & Lib_Name (Project_Id) & ".a";
      Target_File   : aliased String := Target_Dir & Simple_Name;
      Path_Test     : aliased String := "-path";
      Prune_Action  : aliased String := "-prune";
      Or_Symbol     : aliased String := "-o";
      Name_Test     : aliased String := "-name";
      Print_Action  : aliased String := "-print";
      Temp_File : GNAT.OS_Lib.String_Access;
      File : File_Type;
   begin
      Misc.Capture_Output ("/usr/bin/find",
                           (Path_Test'Unchecked_Access,
                            Target_File'Unchecked_Access,
                            Prune_Action'Unchecked_Access,
                            Or_Symbol'Unchecked_Access,
                            Name_Test'Unchecked_Access,
                            Simple_Name'Unchecked_Access,
                            Print_Action'Unchecked_Access), Temp_File);
      Open (File, In_File, Temp_File.all);
      if End_Of_File (File) then
         Dh_Lib.Warning ("unable to find " & Simple_Name & ", not installed.");
      else
         declare
            Found_Path : constant String := Get_Line (File);
         begin
            if End_Of_File (File) then
               Dh_Lib.Create_Path (Target_Dir);
               Dh_Lib.Copy_File (Found_Path, Target_File);
            else
               Dh_Lib.Warning ("many files named " & Simple_Name
                                 & ", none installed.");
            end if;
         end;
      end if;
      Close (File);
      Ada.Directories.Delete_File (Temp_File.all);
      Free (Temp_File);
   exception
      when others =>
         if Is_Open (File) then Close (File); end if;
         if Temp_File /= null then
            Ada.Directories.Delete_File (Temp_File.all);
            Free (Temp_File);
         end if;
         raise;
   end Install_Static_Library;

   function Lib_Name
     (Project_Id : not null Prj.Project_Id)
     return String
   is
   begin
      return Namet.Get_Name_String (Project_Id.all.Library_Name);
   end Lib_Name;

   procedure List_Packages_In_Debian_Control
     (Packages_In_Control : out String_Sets.Set)
   is
      use Ada.Text_IO;
      use GNAT.Regpat;
      Pattern : constant Pattern_Matcher := Compile ("^Package: *([^ ]+)$");
      Matches : Match_Array (0 .. 1);
      File : File_Type;
   begin
      Open (File, In_File, "debian/control");
      while not End_Of_File (File) loop
         declare
            L : constant String := Get_Line (File);
         begin
            Match (Pattern, L, Matches);
            if Matches (0) /= No_Match then
               String_Sets.Insert
                 (Container => Packages_In_Control,
                  New_Item  => L (Matches (1).First .. Matches (1).Last));
            end if;
         end;
      end loop;
      Close (File);
   exception
      when others =>
         if Is_Open (File) then Close (File); end if;
         raise;
   end List_Packages_In_Debian_Control;

   function Package_Containing
     (Pattern : not null GNAT.OS_Lib.String_Access)
     return String
   is
      use Ada.Text_IO;
      use GNAT.OS_Lib;
      Temp_File   : String_Access;
      Return_Code : Integer;
      File        : File_Type;
      Search      : aliased String := "--search";
   begin
      Misc.Capture_Output ("/usr/bin/dpkg-query",
                           (Search'Unchecked_Access,
                            Pattern), Temp_File, Return_Code);
      if Return_Code = 0 then
         Misc.Raise_Subprocess_Error ("/usr/bin/dpkg-query",
                                      (Search'Unchecked_Access,
                                       Pattern),
                                      Return_Code);
      elsif Return_Code = 1 then           --  no match
         Ada.Directories.Delete_File (Temp_File.all);
         Free (Temp_File);
         return "";
      end if;
      Open (File, In_File, Temp_File.all);
      declare
         Buffer     : constant String := Get_Line (File);
         Many_Lines : constant Boolean := not End_Of_File (File);
      begin
         Close (File);
         Ada.Directories.Delete_File (Temp_File.all);
         Free (Temp_File);
         if Many_Lines then
            return "";
         else
            return Buffer (Buffer'First
                             .. Ada.Strings.Fixed.Index (Buffer, ": ") - 1);
         end if;
      end;
   exception
      when others =>
         if Temp_File /= null then
            Ada.Directories.Delete_File (Temp_File.all);
            Free (Temp_File);
         end if;
         raise;
   end Package_Containing;

   procedure Parse_Project
     (Project          : in       Project_Argument_Group;
      Project_Tree_Ref : not null Prj.Project_Tree_Ref;
      Project_Id       :      out Prj.Project_Id)
   is
      use Affectation_Argument_Vectors;
      Insignificant_String_Access : GNAT.OS_Lib.String_Access;
      Unsavory_Project_Node_Id : Prj.Tree.Project_Node_Id;
      Soporific_Project_Node_Tree_Ref : Prj.Tree.Project_Node_Tree_Ref;
      Spiceless_Boolean : Boolean;
      Changes : Environment_Variable_Backups.Backup;
      Aff : Affectation_Argument_Vectors.Cursor;
   begin
      Aff := First (Project.Affectations);
      while Aff /= Affectation_Argument_Vectors.No_Element loop
         Environment_Variable_Backups.Set (Element (Aff).Name,
                                           Element (Aff).Value, Changes);
         Next (Aff);
      end loop;

      --  Dynamic allocation is unavoidable here, as the only way to
      --  deallocate substructures in Prj is Prj.Tree.Free.
      Soporific_Project_Node_Tree_Ref := new Prj.Tree.Project_Node_Tree_Data;
      Prj.Tree.Initialize (Soporific_Project_Node_Tree_Ref);
      Prj.Conf.Parse_Project_And_Apply_Config
        (Main_Project            => Project_Id,
         User_Project_Node       => Unsavory_Project_Node_Id,
         Autoconf_Specified      => True,
         Project_File_Name       => Project.Path,
         Project_Tree            => Project_Tree_Ref,
         Project_Node_Tree       => Soporific_Project_Node_Tree_Ref,
         Packages_To_Check       => Prj.All_Packages,
         Automatically_Generated => Spiceless_Boolean,
         Config_File_Path        => Insignificant_String_Access,
         Normalized_Hostname     => "",
         Flags                   => Prj.Gprbuild_Flags);

      Environment_Variable_Backups.Restore (Changes);
      GNAT.OS_Lib.Free (Insignificant_String_Access);
      Prj.Tree.Free (Soporific_Project_Node_Tree_Ref);
      pragma Unreferenced (Spiceless_Boolean);
      pragma Unreferenced (Unsavory_Project_Node_Id);
   exception
      when others =>
         GNAT.OS_Lib.Free (Insignificant_String_Access);
         Prj.Tree.Free (Soporific_Project_Node_Tree_Ref);
         raise;
   end Parse_Project;

   procedure Process_Library
     (Project_Tree_Ref    : not null Prj.Project_Tree_Ref;
      Project_Id          : not null Prj.Project_Id;
      Packages_In_Control : in       String_Sets.Set)
   is
      Name_Minus  : constant String
        := With_Minus_If_Needed (Lib_Name (Project_Id));
      Dbg_Pattern : constant String := "^(lib" & Lib_Name (Project_Id)
        & "-dbg)$";
      Dev_Pattern : constant String := "^(lib" & Name_Minus
        & "[[:digit:].]+-dev)$";
      Lib_Pattern : constant String := "^lib" & Name_Minus
        & "([[:digit:].]+)$";
      Dbg_Pkg     : constant String
        := Find_Package (Packages_In_Control, Dbg_Pattern);
      Dev_Pkg     : constant String
        := Find_Package (Packages_In_Control, Dev_Pattern);
      Soversion   : constant String
        := Find_Package (Packages_In_Control, Lib_Pattern);
      Lib_Pkg     : constant String := "lib" & Name_Minus & Soversion;
      Soname      : aliased  String := "lib" & Name_Minus & ".so." & Soversion;
      Dev_Pkg_Dir : constant String := "debian/" & Dev_Pkg & '/';
      Lib_Pkg_Dir : constant String := "debian/" & Lib_Pkg & '/';
   begin
      if Dbg_Pkg = "" then
         raise Consistency_Error with
           "No package matching " & Dbg_Pattern & " in debian/control.";
      elsif Dev_Pkg = "" then
         raise Consistency_Error with
           "No package matching " & Dev_Pattern & " in debian/control.";
      elsif Soversion = "" then
         raise Consistency_Error with
           "No package matching " & Lib_Pattern & " in debian/control.";
      end if;
      if String_Sets.Contains (Dh_Lib.Dopackages, Lib_Pkg) then
         Install_Shared_Library (Project_Id, Lib_Pkg_Dir, Soname);
         Run_Dh_Strip_For_Shared_Library (Lib_Pkg    => Lib_Pkg,
                                          Dbg_Pkg    => Dbg_Pkg);
         Append_Trampoline_Lintian_Override (Project_Id  => Project_Id,
                                             Lib_Pkg     => Lib_Pkg,
                                             Lib_Pkg_Dir => Lib_Pkg_Dir,
                                             Soname      => Soname);
         Dh_Lib.Write_Log (Binary_Package => Lib_Pkg);
      end if;
      if String_Sets.Contains (Dh_Lib.Dopackages, Dev_Pkg) then
         Install_Ali_Files (Project_Id, Dev_Pkg_Dir);
         Install_Sources (Project_Tree_Ref, Project_Id, Dev_Pkg_Dir);
         Install_Shared_Object_Symbolic_Link
           (Project_Id  => Project_Id,
            Dev_Pkg_Dir => Dev_Pkg_Dir,
            Soname      => Soname'Unchecked_Access);
         Install_Static_Library (Project_Id, Dev_Pkg_Dir);
         Install_Project (Project_Tree_Ref, Project_Id, Dev_Pkg_Dir);
         Run_Dh_Strip_Excluding_Static_Archive (Project_Id, Dev_Pkg);
         Dh_Lib.Addsubstvar (Dev_Pkg, "ada:Depends", "gnat");
         Dh_Lib.Addsubstvar (Dev_Pkg, "ada:Depends",
                             "gnat-" & Gnatgcc_Version);
         Dh_Lib.Addsubstvar (Dev_Pkg, "ada:Depends", Lib_Pkg,
                             "= ${binary:Version}");
         Add_Substitution_Variables_For_Withed_Projects (Project_Id, Dev_Pkg,
                                                         Packages_In_Control);
         Dh_Lib.Write_Log (Binary_Package => Dev_Pkg);
      end if;
      if String_Sets.Contains (Dh_Lib.Dopackages, Dbg_Pkg) then
         Run_Dh_Installdocs (Pkg      => Dbg_Pkg,
                             Link_Doc => Lib_Pkg);
         Dh_Lib.Addsubstvar (Dbg_Pkg, "ada:Depends", Lib_Pkg,
                             "= ${binary:Version}");
         Dh_Lib.Addsubstvar (Dbg_Pkg, "ada:Recommends", Dev_Pkg,
                             "= ${binary:Version}");
         Dh_Lib.Addsubstvar (Dbg_Pkg, "ada:Suggests", "gnat");
         Dh_Lib.Write_Log (Binary_Package => Dbg_Pkg);
      end if;
   end Process_Library;

   procedure Run_Dh_Installdocs
     (Pkg      : in String;
      Link_Doc : in String)
   is
      Pkg_Arg      : aliased String := "--package=" & Pkg;
      Link_Doc_Arg : aliased String := "--link-doc=" & Link_Doc;
   begin
      Dh_Lib.Do_It ("/usr/bin/dh_installdocs",
                    (Pkg_Arg'Unchecked_Access,
                     Link_Doc_Arg'Unchecked_Access));
   end Run_Dh_Installdocs;

   procedure Run_Dh_Strip_Excluding_Static_Archive
     (Project_Id : not null Prj.Project_Id;
      Dev_Pkg    : in       String)
   is
      Pkg_Equals     : aliased String := "--package=" & Dev_Pkg;
      Exclude_Static : aliased String := "--exclude=" & Lib_Install_Dir
        & "lib" & Lib_Name (Project_Id) & ".a";
   begin
      Dh_Lib.Do_It ("/usr/bin/dh_strip",
                    (Pkg_Equals'Unchecked_Access,
                     Exclude_Static'Unchecked_Access));
   end Run_Dh_Strip_Excluding_Static_Archive;

   procedure Run_Dh_Strip_For_Shared_Library
     (Lib_Pkg : in String;
      Dbg_Pkg : in String)
   is
      Pkg_Equals_Lib : aliased String := "--package=" & Lib_Pkg;
      Dbg_To_Pkg     : aliased String := "--dbg-package=" & Dbg_Pkg;
   begin
      if String_Sets.Contains (Dh_Lib.Dopackages, Dbg_Pkg) then
         Dh_Lib.Do_It ("/usr/bin/dh_strip",
                       (Pkg_Equals_Lib'Unchecked_Access,
                        Dbg_To_Pkg'Unchecked_Access));
      else
         Dh_Lib.Warning ("Package " & Dbg_Pkg
                           & " is not processed by debhelper,"
                           & " unable to strip shared library");
      end if;
   end Run_Dh_Strip_For_Shared_Library;

   procedure Parse_Arguments
     (Projects  : in out Project_Argument_Group_Vectors.Vector;
      Processed :    out Dh_Lib.Command_Line_Argument_Set)
   is
      procedure Check_At_End;
      procedure Process_Argument (Argument   : in     String;
                                  Recognized :    out Boolean);
      procedure Process_Line (Buffer : in String);
      Affectations : Affectation_Argument_Vectors.Vector;

      procedure Check_At_End is
      begin
         if not Affectation_Argument_Vectors.Is_Empty (Affectations) then
            raise Usage_Error
              with " affectations must come before the project";
         end if;
      end Check_At_End;

      procedure Process_Argument (Argument   : in     String;
                                  Recognized :    out Boolean) is
         use Ada.Strings.Fixed;
         Eq : Natural;
      begin
         if Tail (Argument, 4) = ".gpr" then
            Project_Argument_Group_Vectors.Append
              (Projects, (Path         => Argument,
                          Path_Length  => Argument'Length,
                          Affectations => Affectations));
            Affectation_Argument_Vectors.Clear (Affectations);
            Recognized := True;
         else
            Eq := Index (Argument, "=");
            if Eq /= 0 then
               Affectation_Argument_Vectors.Append
                 (Affectations,
                  (Name         => Argument (Argument'First .. Eq - 1),
                   Name_Length  => Eq - Argument'First,
                   Value        => Argument (Eq + 1 .. Argument'Last),
                   Value_Length => Argument'Last - Eq));
               Recognized := True;
            else
               Recognized := False;
            end if;
         end if;
      end Process_Argument;

      procedure Process_Line (Buffer : in String) is
         use Ada.Strings.Fixed;
         First      : Natural;
         Last       : Natural := Buffer'First - 1;
         Recognized : Boolean;
      begin
         if Buffer'Length > 0 and then Buffer (Buffer'First) /= '#' then
            while Last < Buffer'Last loop
               First := Index (Buffer, Space_Character_Set, Last + 1,
                               Ada.Strings.Outside);
               exit when First = 0;
               Last := Index (Buffer, Space_Character_Set, First,
                              Ada.Strings.Inside);
               if Last = 0 then
                  Last := Buffer'Last;
               else
                  Last := Last - 1;
               end if;
               Process_Argument (Buffer (First .. Last), Recognized);
               if not Recognized then
                  raise Usage_Error
                    with Configuration_File & ": unrecognized argument "
                    & Buffer (First .. Last);
               end if;
            end loop;
         end if;
      end Process_Line;

      use Ada.Text_IO;
      File : File_Type;
   begin
      --  Command line
      for I in Processed'Range loop
         Process_Argument (Ada.Command_Line.Argument (I), Processed (I));
      end loop;
      Check_At_End;
      --  debian/ada_libraries
      if Ada.Directories.Exists (Configuration_File) then
         Open (File, In_File, Configuration_File);
         while not End_Of_File (File) loop
            Process_Line (Get_Line (File));
         end loop;
         Close (File);
         Check_At_End;
      end if;
   exception
      when others =>
         if Is_Open (File) then Close (File); end if;
         raise;
   end Parse_Arguments;

   function With_Minus_If_Needed
     (Name : in String)
     return String
   is
   begin
      if Ada.Characters.Handling.Is_Decimal_Digit (Name (Name'Last)) then
         return Name & '-';
      else
         return Name;
      end if;
   end With_Minus_If_Needed;

   ----------------------------------------------------------------------
   --  Main body

   use Project_Argument_Group_Vectors;
   Packages_In_Control : String_Sets.Set;
   Projects            : Vector;
   Project             : Cursor;
   Processed           : Dh_Lib.Command_Line_Argument_Set;
   Project_Tree_Data   : aliased Prj.Project_Tree_Data;
   Project_Id          : Prj.Project_Id;
begin
   Parse_Arguments (Projects, Processed);
   Dh_Lib.Init (Processed);

   --  Now, usage error should have been detected, we may exit if we
   --  are certain that there is nothing to do.
   if String_Sets.Is_Empty (Dh_Lib.Dopackages) or Is_Empty (Projects) then
      return;
   end if;

   --  Initialize list of all packages.
   List_Packages_In_Debian_Control (Packages_In_Control);

   --  Initialize Prj dependencies.
   Namet.Initialize;
   Snames.Initialize;
   Csets.Initialize;
   Prj.Initialize (Project_Tree_Data'Unchecked_Access);
   Prj.Pars.Set_Verbosity (Prj.Default);

   Project := First (Projects);
   while Project /= No_Element loop
      Parse_Project (Element (Project), Project_Tree_Data'Unchecked_Access,
                     Project_Id);
      if Project_Id = Prj.No_Project then
         raise Project_Parse_Error with
           Element (Project).Path & ": parsing failed";
      elsif not Project_Id.all.Library then
         raise Consistency_Error with
           Element (Project).Path & ": not a library project";
      elsif Project_Id.all.Library_Kind /= Prj.Dynamic then
         raise Consistency_Error with
           Element (Project).Path & ": not a dynamic library";
      elsif Project_Id.all.Externally_Built then
         raise Consistency_Error with
           Element (Project).Path & ": marked as externally build";
      elsif Project_Id.all.Need_To_Build_Lib then
         raise Consistency_Error with
           Element (Project).Path & ": not compiled yet";
      end if;

      Process_Library (Project_Tree_Data'Unchecked_Access, Project_Id,
                       Packages_In_Control);

      Next (Project);
   end loop;

   --  Well, structures pointed by Project_Tree_Data are not
   --  deallocated at this step, but executions ends anyway.
end Dh_Ada_Library;
